# OpenTrafficShaper Linux tc traffic shaping
# Copyright (C) 2007-2023, AllWorldIT
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.



package opentrafficshaper::plugins::tc;

use strict;
use warnings;

use JSON;
use List::Util qw(min max);
use POE qw(
	Wheel::Run Filter::Line
);

use awitpt::util qw(
	toHex
);
use opentrafficshaper::constants;
use opentrafficshaper::logger;
use opentrafficshaper::plugins::configmanager qw(
	getPool
	getPoolAttribute
	setPoolAttribute
	removePoolAttribute
	getPoolTxInterface
	getPoolRxInterface
	setPoolShaperState
	unsetPoolShaperState
	getPoolShaperState

	getEffectivePool

	getPoolMember
	setPoolMemberAttribute
	getPoolMemberAttribute
	removePoolMemberAttribute
	getPoolMemberMatchPriority
	setPoolMemberShaperState
	unsetPoolMemberShaperState
	getPoolMemberShaperState

	getTrafficClassPriority

	getAllTrafficClasses

	getInterface
	getInterfaceGroup
	getInterfaceGroups
	getInterfaces
	getInterfaceDefaultPool
	getEffectiveInterfaceTrafficClass2
	isInterfaceTrafficClassValid
	setInterfaceTrafficClassShaperState
	unsetInterfaceTrafficClassShaperState
);


# Exporter stuff
require Exporter;
our (@ISA,@EXPORT,@EXPORT_OK);
@ISA = qw(Exporter);
@EXPORT = qw(
);
@EXPORT_OK = qw(
);

use constant {
	VERSION => '1.0.1',

	# 5% of a link can be used for very high priority traffic
	PROTO_RATE_LIMIT => 5,
	PROTO_RATE_BURST_MIN => 16, # With a minimum burst of 8KiB
	PROTO_RATE_BURST_MAXM => 1.5, # Multiplier for burst min to get to burst max

	# High priority traffic gets the first 20% of the bandidth to itself
	PRIO_RATE_LIMIT => 20,
	PRIO_RATE_BURST_MIN => 32, # With a minimum burst of 40KiB
	PRIO_RATE_BURST_MAXM => 1.5, # Multiplier for burst min to get to burst max

	TC_ROOT_CLASS => 1,
};


# Plugin info
our $pluginInfo = {
	Name => "Linux tc Interface",
	Version => VERSION,

	Init => \&plugin_init,
	Start => \&plugin_start,
};


# Our globals
my $globals;
# Copy of system logger
my $logger;

# Our configuration
my $config = {
	'ip_protocol' => "ip",
	'iphdr_offset' => 0,
};

#
# TASK QUEUE
#
# $globals->{'TaskQueue'}

#
# TC CLASSES & FILTERS
#
# $globals->{'TcClasses'}
# $globals->{'TcFilterMappings'}
# $globals->{'TcFilters'}


# Initialize plugin
sub plugin_init
{
	my $system = shift;


	# Setup our environment
	$logger = $system->{'logger'};

	$logger->log(LOG_NOTICE,"[TC] OpenTrafficShaper tc Integration v%s - Copyright (c) 2007-2023, AllWorldIT",VERSION);

	# Initialize
	$globals->{'TaskQueue'} = [ ];
	$globals->{'TcClasses'} = { };
	$globals->{'TcFilterMappings'} = { };
	$globals->{'TcFilters'} = { };

	# Grab some of our config we need
	if (defined(my $proto = $system->{'file.config'}->{'plugin.tc'}->{'protocol'})) {
		$logger->log(LOG_INFO,"[TC] Set protocol to '%s'",$proto);
		$config->{'ip_protocol'} = $proto;
	}
	if (defined(my $offset = $system->{'file.config'}->{'plugin.tc'}->{'iphdr_offset'})) {
		$logger->log(LOG_INFO,"[TC] Set IP header offset to '%s'",$offset);
		$config->{'iphdr_offset'} = $offset;
	}


	# We going to queue the initialization in plugin initialization so nothing at all can come before us
	my $changeSet = TC::ChangeSet->new();

	# Loop with protocols
	for my $ipv ("", "6") {

		#
		# Traffic Shaping
		#

		# First the Cleanup
		#
		foreach my $interfaceID (getInterfaces()) {
			my $interface = getInterface($interfaceID);
			# Flush inteface chain
			$changeSet->add([
				"/sbin/ip${ipv}tables",
					'-t','mangle',
					'-F',"ots-tcfor-$interfaceID",
			]);
			# Delete jump
			$changeSet->add([
				"/sbin/ip${ipv}tables",
					'-t','mangle',
					'-D','ots-tcfor',
					'-o',$interface->{'Device'},
					'-j',"ots-tcfor-$interfaceID",
			]);
			# Delete interface chain
			$changeSet->add([
				"/sbin/ip${ipv}tables",
					'-t','mangle',
					'-X',"ots-tcfor-$interfaceID",
			]);
		}
		# Flush ots-tcfor
		$changeSet->add([
			"/sbin/ip${ipv}tables",
				'-t','mangle',
				'-F','ots-tcfor',
		]);
		# Delete jump to ots-tcfor
		$changeSet->add([
			"/sbin/ip${ipv}tables",
				'-t','mangle',
				'-D','FORWARD',
				'-j','ots-tcfor',
		]);
		# Delete ots-tcfor
		$changeSet->add([
			"/sbin/ip${ipv}tables",
				'-t','mangle',
				'-X','ots-tcfor',
		]);

		# Then the setup
		#

		# Create ots-tcfor
		$changeSet->add([
			"/sbin/ip${ipv}tables",
				'-t','mangle',
				'-N','ots-tcfor',
		]);
		# Add jump to ots-tcfor
		$changeSet->add([
			"/sbin/ip${ipv}tables",
				'-t','mangle',
				'-I','FORWARD',
				'-j','ots-tcfor',
		]);
		# Add interface chains
		foreach my $interfaceID (getInterfaces()) {
			my $interface = getInterface($interfaceID);
			# Create ots-tcfor
			$changeSet->add([
				"/sbin/ip${ipv}tables",
					'-t','mangle',
					'-N',"ots-tcfor-$interfaceID",
			]);
			# Add jump to ots-tcfor
			$changeSet->add([
				"/sbin/ip${ipv}tables",
					'-t','mangle',
					'-A','ots-tcfor',
					'-o',$interface->{'Device'},
					'-j',"ots-tcfor-$interfaceID",
			]);
		}

		#
		# NAT
		#

		# First the Cleanup
		#
		foreach my $interfaceGroupName (getInterfaceGroups()) {
			my $interfaceGroup = getInterfaceGroup($interfaceGroupName);
			my $txInterfaceID = $interfaceGroup->{'TxInterface'};
			my $rxInterfaceID = $interfaceGroup->{'RxInterface'};
			my $txInterface = getInterface($txInterfaceID);
			my $rxInterface = getInterface($rxInterfaceID);

			# Flush inteface chains
			$changeSet->add([
				"/sbin/ip${ipv}tables",
					'-t','nat',
					'-F',"ots-snat-$txInterfaceID",
			]);
			$changeSet->add([
				"/sbin/ip${ipv}tables",
					'-t','nat',
					'-F',"ots-dnat-$rxInterfaceID",
			]);
			# Delete interface chains
			$changeSet->add([
				"/sbin/ip${ipv}tables",
					'-t','nat',
					'-X',"ots-snat-$txInterfaceID",
			]);
			$changeSet->add([
				"/sbin/ip${ipv}tables",
					'-t','nat',
					'-X',"ots-dnat-$rxInterfaceID",
			]);
		}
		# Flush ots-snat and ots-dnat
		$changeSet->add([
			"/sbin/ip${ipv}tables",
				'-t','nat',
				'-F','ots-snat',
		]);
		$changeSet->add([
			"/sbin/ip${ipv}tables",
				'-t','nat',
				'-F','ots-dnat',
		]);
		# Delete jump to ots-snat and ots-dnat
		$changeSet->add([
			"/sbin/ip${ipv}tables",
				'-t','nat',
				'-D','POSTROUTING',
				'-j','ots-snat',
		]);
		$changeSet->add([
			"/sbin/ip${ipv}tables",
				'-t','nat',
				'-D','OUTPUT',
				'-j','ots-dnat',
		]);
		$changeSet->add([
			"/sbin/ip${ipv}tables",
				'-t','nat',
				'-D','PREROUTING',
				'-j','ots-dnat',
		]);
		# Delete ots-tcfor
		$changeSet->add([
			"/sbin/ip${ipv}tables",
				'-t','nat',
				'-X','ots-snat',
		]);
		$changeSet->add([
			"/sbin/ip${ipv}tables",
				'-t','nat',
				'-X','ots-dnat',
		]);

		# Then the setup
		#

		# Create ots-snat and ots-dnat
		$changeSet->add([
			"/sbin/ip${ipv}tables",
				'-t','nat',
				'-N','ots-snat',
		]);
		$changeSet->add([
			"/sbin/ip${ipv}tables",
				'-t','nat',
				'-N','ots-dnat',
		]);
		# Add jump to ots-snat and ots-dnat
		$changeSet->add([
			"/sbin/ip${ipv}tables",
				'-t','nat',
				'-I','POSTROUTING',
				'-j','ots-snat',
		]);
		$changeSet->add([
			"/sbin/ip${ipv}tables",
				'-t','nat',
				'-I','OUTPUT',
				'-j','ots-dnat',
		]);
		$changeSet->add([
			"/sbin/ip${ipv}tables",
				'-t','nat',
				'-A','PREROUTING',
				'-j','ots-dnat',
		]);
		foreach my $interfaceGroupName (getInterfaceGroups()) {
			my $interfaceGroup = getInterfaceGroup($interfaceGroupName);
			my $txInterfaceID = $interfaceGroup->{'TxInterface'};
			my $rxInterfaceID = $interfaceGroup->{'RxInterface'};
			my $txInterface = getInterface($txInterfaceID);
			my $rxInterface = getInterface($rxInterfaceID);

			# Flush inteface chain
			$changeSet->add([
				"/sbin/ip${ipv}tables",
					'-t','nat',
					'-N',"ots-snat-$txInterfaceID",
			]);
			$changeSet->add([
				"/sbin/ip${ipv}tables",
					'-t','nat',
					'-N',"ots-dnat-$rxInterfaceID",
			]);
			# Add interface chains
			$changeSet->add([
				"/sbin/ip${ipv}tables",
					'-t','nat',
					'-A','ots-snat',
					'-o',$rxInterface->{'Device'},
					'-j',"ots-snat-$txInterfaceID",
			]);
			$changeSet->add([
				"/sbin/ip${ipv}tables",
					'-t','nat',
					'-A','ots-snat',
					'-o',$txInterface->{'Device'},
					'-j',"ots-snat-$txInterfaceID",
			]);
			$changeSet->add([
				"/sbin/ip${ipv}tables",
					'-t','nat',
					'-A','ots-dnat',
					'-i',$rxInterface->{'Device'},
					'-j',"ots-dnat-$rxInterfaceID",
			]);
			$changeSet->add([
				"/sbin/ip${ipv}tables",
					'-t','nat',
					'-A','ots-dnat',
					'-i',$txInterface->{'Device'},
					'-j',"ots-dnat-$rxInterfaceID",
			]);
		}
	}

	# Loop with the configured interfaces and initialize them
	foreach my $interfaceID (getInterfaces()) {
		my $interface = getInterface($interfaceID);
		# Initialize interface
		$logger->log(LOG_INFO,"[TC] Queuing tasks to initialize '%s'",$interface->{'Device'});
		_tc_iface_init($changeSet,$interfaceID);
	}
	_task_add_to_queue($changeSet);


	# This session is our main session, its alias is "shaper"
	POE::Session->create(
		inline_states => {
			_start => \&_session_start,
			_stop => \&_session_stop,

			class_change => \&_session_class_change,

			pool_add => \&_session_pool_add,
			pool_remove => \&_session_pool_remove,
			pool_change => \&_session_pool_change,

			poolmember_add => \&_session_poolmember_add,
			poolmember_remove => \&_session_poolmember_remove,
		}
	);

	# This is our session for communicating directly with tc, its alias is _tc
	POE::Session->create(
		inline_states => {
			_start => \&_task_session_start,
			_stop => sub { },
			# Signals
			_SIGCHLD => \&_task_SIGCHLD,
			_SIGINT => \&_task_SIGINT,

			# Public'ish
			queue => \&_task_queue,

			# Internal
			_task_child_stdout => \&_task_child_stdout,
			_task_child_stderr => \&_task_child_stderr,
			_task_child_stdin => \&_task_child_stdin,
			_task_child_close => \&_task_child_close,
			_task_child_error => \&_task_child_error,
			_task_run_next => \&_task_run_next,
		}
	);

	return 1;
}



# Start the plugin
sub plugin_start
{
	$logger->log(LOG_INFO,"[TC] Started");
}



# Initialize this plugins main POE session
sub _session_start
{
	my ($kernel,$heap) = @_[KERNEL,HEAP];


	# Set our alias
	$kernel->alias_set("shaper");

	$logger->log(LOG_DEBUG,"[TC] Initialized");
}



# Initialize this plugins main POE session
sub _session_stop
{
	my ($kernel,$heap) = @_[KERNEL,HEAP];


	# Remove our alias
	$kernel->alias_remove("shaper");

	# Blow away data
	$globals = undef;

	$logger->log(LOG_DEBUG,"[TC] Shutdown");

	$logger = undef;
}



# Event handler for changing a class
sub _session_class_change
{
	my ($kernel, $interfaceTrafficClassID) = @_[KERNEL, ARG0, ARG1];


	# Grab our effective class
	my $effectiveInterfaceTrafficClass = getEffectiveInterfaceTrafficClass2($interfaceTrafficClassID);

	# Grab interface ID
	my $interfaceID = $effectiveInterfaceTrafficClass->{'InterfaceID'};
	# Grab interface from config manager
	my $interface = getInterface($interfaceID);

	# Grab traffic class ID
	my $trafficClassID = $effectiveInterfaceTrafficClass->{'TrafficClassID'};

	$logger->log(LOG_INFO,"[TC] Processing interface class changes for '%s' traffic class ID '%s'",
			$interface->{'Device'},
			$trafficClassID
	);

	# Grab tc interface
	my $tcInterface = $globals->{'Interfaces'}->{$interfaceID};
	# Grab interface traffic class
	my $interfaceTrafficClass = $tcInterface->{'TrafficClasses'}->{$trafficClassID};

	# Grab the traffic class
	my $majorTcClass = $tcInterface->{'TcClass'};
	my $minorTcClass = $interfaceTrafficClass->{"TcClass"};

	# Generate changeset
	my $changeSet = TC::ChangeSet->new();

	# If we're a normal class we are treated differently than if we're a main/root class below (interface main speed)
	if ($minorTcClass > 1) {
		_tc_class_change($changeSet,$interfaceID,$majorTcClass,"",$minorTcClass,
				$effectiveInterfaceTrafficClass->{'CIR'},
				$effectiveInterfaceTrafficClass->{'Limit'}
		);
	# XXX: This will be the actual interface, we set limit and burst to the same
	} else {
		_tc_class_change($changeSet,$interfaceID,TC_ROOT_CLASS,"",$minorTcClass,$effectiveInterfaceTrafficClass->{'Limit'});
	}

	# Post changeset
	$kernel->post("_tc" => "queue" => $changeSet);

	# Mark as live
	unsetInterfaceTrafficClassShaperState($interfaceTrafficClassID,SHAPER_NOTLIVE|SHAPER_PENDING);
	setInterfaceTrafficClassShaperState($interfaceTrafficClassID,SHAPER_LIVE);
}



# Event handler for adding a pool
sub _session_pool_add
{
	my ($kernel,$heap,$pid) = @_[KERNEL, HEAP, ARG0];


	# Grab pool
	my $pool;
	if (!defined($pool = getPool($pid))) {
		$logger->log(LOG_ERR,"[TC] Shaper 'add' event with non existing pool '%s'",$pid);
		return;
	}

	$logger->log(LOG_INFO,"[TC] Add pool '%s' [%s] to interface group '%s'",
			$pool->{'Name'},
			$pool->{'ID'},
			$pool->{'InterfaceGroupID'},
	);

	# Grab our effective pool
	my $effectivePool = getEffectivePool($pool->{'ID'});

	my $changeSet = TC::ChangeSet->new();

	# Grab some things we need from the main pool
	my $txInterfaceID = getPoolTxInterface($pool->{'ID'});
	my $rxInterfaceID = getPoolRxInterface($pool->{'ID'});

	# Grab effective config
	my $trafficClassID = $effectivePool->{'TrafficClassID'};
	my $txCIR = $effectivePool->{'TxCIR'};
	my $txLimit = $effectivePool->{'TxLimit'};
	my $rxCIR = $effectivePool->{'RxCIR'};
	my $rxLimit = $effectivePool->{'RxLimit'};
	my $trafficPriority = getTrafficClassPriority($effectivePool->{'TrafficClassID'});

	# Get the Tx traffic classes TC class
	my $tcClass_TxTrafficClass = _getTcClassFromTrafficClassID($txInterfaceID,$trafficClassID);
	# Generate our pools Tx TC class
	my $tcClass_TxPool = _reserveMinorTcClassByPoolID($txInterfaceID,$pool->{'ID'});
	# Add the main Tx TC class for this pool
	_tc_class_add($changeSet,$txInterfaceID,TC_ROOT_CLASS,$tcClass_TxTrafficClass,$tcClass_TxPool,$txCIR,
			$txLimit,$trafficPriority
	);
	# Add Tx TC optimizations
	_tc_class_optimize($changeSet,$txInterfaceID,$tcClass_TxPool,$txCIR);
	# Set Tx TC class
	setPoolAttribute($pool->{'ID'},'tc.txclass',$tcClass_TxPool);

	# Get the Rx traffic classes TC class
	my $tcClass_RxTrafficClass = _getTcClassFromTrafficClassID($rxInterfaceID,$trafficClassID);
	# Generate our pools Rx TC class
	my $tcClass_RxPool = _reserveMinorTcClassByPoolID($rxInterfaceID,$pool->{'ID'});
	# Add the main Rx TC class for this pool
	_tc_class_add($changeSet,$rxInterfaceID,TC_ROOT_CLASS,$tcClass_RxTrafficClass,$tcClass_RxPool,$rxCIR,
			$rxLimit,$trafficPriority
	);
	# Add Rx TC optimizations
	_tc_class_optimize($changeSet,$rxInterfaceID,$tcClass_RxPool,$rxCIR);
	# Set Rx TC
	setPoolAttribute($pool->{'ID'},'tc.rxclass',$tcClass_RxPool);

	# Post changeset
	$kernel->post("_tc" => "queue" => $changeSet);

	# Set current live values
	setPoolAttribute($pool->{'ID'},'shaper.live.ClassID',$trafficClassID);
	setPoolAttribute($pool->{'ID'},'shaper.live.TxCIR',$txCIR);
	setPoolAttribute($pool->{'ID'},'shaper.live.TxLimit',$txLimit);
	setPoolAttribute($pool->{'ID'},'shaper.live.RxCIR',$rxCIR);
	setPoolAttribute($pool->{'ID'},'shaper.live.RxLimit',$rxLimit);

	# Mark as live
	unsetPoolShaperState($pool->{'ID'},SHAPER_NOTLIVE|SHAPER_PENDING);
	setPoolShaperState($pool->{'ID'},SHAPER_LIVE);
}



# Event handler for removing a pool
sub _session_pool_remove
{
	my ($kernel, $pid) = @_[KERNEL, ARG0];


	my $changeSet = TC::ChangeSet->new();

	# Pull in pool
	my $pool;
	if (!defined($pool = getPool($pid))) {
		$logger->log(LOG_ERR,"[TC] Shaper 'remove' event with non existing pool '%s'",$pid);
		return;
	}

	# Make sure its not NOTLIVE
	if (getPoolShaperState($pid) & SHAPER_NOTLIVE) {
		$logger->log(LOG_WARN,"[TC] Ignoring remove for pool '%s' [%s]",
				$pool->{'Name'},
				$pool->{'ID'}
		);
		return;
	}

	$logger->log(LOG_INFO,"[TC] Removing pool '%s' [%s]",
			$pool->{'Name'},
			$pool->{'ID'}
	);

	# Grab our interfaces
	my $txInterfaceID = getPoolTxInterface($pool->{'ID'});
	my $rxInterfaceID = getPoolRxInterface($pool->{'ID'});
	# Grab the traffic class from the pool
	my $txPoolTcClass = getPoolAttribute($pool->{'ID'},'tc.txclass');
	my $rxPoolTcClass = getPoolAttribute($pool->{'ID'},'tc.rxclass');

	# Grab current class ID
	my $trafficClassID = getPoolAttribute($pool->{'ID'},'shaper.live.ClassID');
	# Grab our minor classes
	my $txTrafficClassTcClass = _getTcClassFromTrafficClassID($txInterfaceID,$trafficClassID);
	my $rxTrafficClassTcClass = _getTcClassFromTrafficClassID($rxInterfaceID,$trafficClassID);

	my $txInterface = getInterface($txInterfaceID);
	my $rxInterface = getInterface($rxInterfaceID);

	# Clear up the class
	$changeSet->add([
			'/sbin/tc','class','del',
				'dev',$txInterface->{'Device'},
				'parent',"1:$txTrafficClassTcClass",
				'classid',"1:$txPoolTcClass",
	]);
	$changeSet->add([
			'/sbin/tc','class','del',
				'dev',$rxInterface->{'Device'},
				'parent',"1:$rxTrafficClassTcClass",
				'classid',"1:$rxPoolTcClass",
	]);

	# And recycle the classs
	_disposePoolTcClass($txInterface->{'Device'},$txPoolTcClass);
	_disposePoolTcClass($rxInterface->{'Device'},$rxPoolTcClass);

	_disposePrioTcClass($txInterface->{'Device'},$txPoolTcClass);
	_disposePrioTcClass($rxInterface->{'Device'},$rxPoolTcClass);

	# Post changeset
	$kernel->post("_tc" => "queue" => $changeSet);

	# Cleanup attributes
	removePoolAttribute($pool->{'ID'},'tc.txclass');
	removePoolAttribute($pool->{'ID'},'tc.rxclass');

	removePoolAttribute($pool->{'ID'},'shaper.live.ClassID');
	removePoolAttribute($pool->{'ID'},'shaper.live.TxCIR');
	removePoolAttribute($pool->{'ID'},'shaper.live.TxLimit');
	removePoolAttribute($pool->{'ID'},'shaper.live.RxCIR');
	removePoolAttribute($pool->{'ID'},'shaper.live.RxLimit');

	# Mark as not live
	unsetPoolShaperState($pool->{'ID'},SHAPER_LIVE|SHAPER_PENDING);
	setPoolShaperState($pool->{'ID'},SHAPER_NOTLIVE);
}



## Event handler for changing a pool
sub _session_pool_change
{
	my ($kernel, $pid) = @_[KERNEL, ARG0];


	# Grab pool
	my $pool = getPool($pid);

	$logger->log(LOG_INFO,"[TC] Processing changes for '%s' [%s]",$pool->{'Name'},$pool->{'ID'});

	# Grab our effective pool
	my $effectivePool = getEffectivePool($pool->{'ID'});

	# Grab our interfaces
	my $txInterfaceID = getPoolTxInterface($pool->{'ID'});
	my $rxInterfaceID = getPoolRxInterface($pool->{'ID'});
	# Grab the traffic class from the pool
	my $txPoolTcClass = getPoolAttribute($pool->{'ID'},'tc.txclass');
	my $rxPoolTcClass = getPoolAttribute($pool->{'ID'},'tc.rxclass');

	# Grab effective config
	my $trafficClassID = $effectivePool->{'TrafficClassID'};
	my $txCIR = $effectivePool->{'TxCIR'};
	my $txLimit = $effectivePool->{'TxLimit'};
	my $rxCIR = $effectivePool->{'RxCIR'};
	my $rxLimit = $effectivePool->{'RxLimit'};
	my $trafficPriority = getTrafficClassPriority($trafficClassID);

	# Grab our minor classes
	my $txTrafficClassTcClass = _getTcClassFromTrafficClassID($txInterfaceID,$trafficClassID);
	my $rxTrafficClassTcClass = _getTcClassFromTrafficClassID($rxInterfaceID,$trafficClassID);

	# Generate changeset
	my $changeSet = TC::ChangeSet->new();

	_tc_class_change($changeSet,$txInterfaceID,TC_ROOT_CLASS,$txTrafficClassTcClass,$txPoolTcClass,$txCIR,
			$txLimit,$trafficPriority);
	_tc_class_change($changeSet,$rxInterfaceID,TC_ROOT_CLASS,$rxTrafficClassTcClass,$rxPoolTcClass,$rxCIR,
			$rxLimit,$trafficPriority);

	# Post changeset
	$kernel->post("_tc" => "queue" => $changeSet);

	setPoolAttribute($pool->{'ID'},'shaper.live.ClassID',$trafficClassID);
	setPoolAttribute($pool->{'ID'},'shaper.live.TxCIR',$txCIR);
	setPoolAttribute($pool->{'ID'},'shaper.live.TxLimit',$txLimit);
	setPoolAttribute($pool->{'ID'},'shaper.live.RxCIR',$rxCIR);
	setPoolAttribute($pool->{'ID'},'shaper.live.RxLimit',$rxLimit);

	# Mark as live
	unsetPoolShaperState($pool->{'ID'},SHAPER_NOTLIVE|SHAPER_PENDING);
	setPoolShaperState($pool->{'ID'},SHAPER_LIVE);
}



# Event handler for adding a pool member
sub _session_poolmember_add
{
	my ($kernel,$heap,$pmid) = @_[KERNEL, HEAP, ARG0];


	# Grab pool
	my $poolMember;
	if (!defined($poolMember = getPoolMember($pmid))) {
		$logger->log(LOG_ERR,"[TC] Shaper 'add' event with non existing pool member '%s'",$pmid);
		return;
	}

	# Grab the pool members associated pool
	my $pool;
	if (!defined($pool = getPool($poolMember->{'PoolID'}))) {
		$logger->log(LOG_ERR,"[TC] Shaper 'poolmember_add' event with invalid PoolID");
		return;
	}

	$logger->log(LOG_INFO,"[TC] Add pool member '%s' [%s] with IP '%s', NAT '%s' (inbound: %s) to pool '%s' [%s]",
			$poolMember->{'Username'},
			$poolMember->{'ID'},
			$poolMember->{'IPAddress'},
			$poolMember->{'IPNATAddress'} // "",
			$poolMember->{'IPNATInbound'} // "",
			$pool->{'Name'},
			$pool->{'ID'}
	);

	my $changeSet = TC::ChangeSet->new();

	my $txInterfaceID = getPoolTxInterface($pool->{'ID'});
	my $rxInterfaceID = getPoolRxInterface($pool->{'ID'});

	my $rxPoolTcClass = getPoolAttribute($pool->{'ID'},'tc.rxclass');
	my $txPoolTcClass = getPoolAttribute($pool->{'ID'},'tc.txclass');

	# Check what IP version we're dealing with
	my $ipv = "";
	if ($poolMember->{'IPAddress'} =~ /:/) {
		$ipv = "6";
	}
	# Add traffic classification
	$changeSet->add([
			"/sbin/ip${ipv}tables",
				'-t','mangle',
				'-A',"ots-tcfor-$rxInterfaceID",
				'-s',$poolMember->{'IPAddress'},
				'-m','comment',
					'--comment', "pool: ".$pool->{'Name'}.", member: ".$poolMember->{'Username'},
				'-j','CLASSIFY',
					'--set-class',"1:$rxPoolTcClass",
	]);
	$changeSet->add([
			"/sbin/ip${ipv}tables",
				'-t','mangle',
				'-A',"ots-tcfor-$txInterfaceID",
				'-d',$poolMember->{'IPAddress'},
				'-m','comment',
					'--comment', "pool: ".$pool->{'Name'}.", member: ".$poolMember->{'Username'},
				'-j','CLASSIFY',
					'--set-class',"1:$txPoolTcClass",
	]);
	# Add NAT if we have any
	if (defined($poolMember->{'IPNATAddress'}) && defined($poolMember->{'IPNATAddress'}) ne "") {
		$changeSet->add([
				"/sbin/ip${ipv}tables",
					'-t','nat',
					'-A',"ots-snat-$txInterfaceID",
					'-s',$poolMember->{'IPAddress'},
					'-m','comment',
						'--comment', "pool: ".$pool->{'Name'}.", member: ".$poolMember->{'Username'},
					'-j','SNAT',
						'--to-source', $poolMember->{'IPNATAddress'},
		]);
		# If we're dealing with IPv4, clear the connection tracking
		if ($ipv eq "") {
			$changeSet->add(["/sbin/conntrack",'-D','-s',$poolMember->{'IPAddress'}]);
		}
		if (defined($poolMember->{'IPNATInbound'}) && $poolMember->{'IPNATInbound'} eq "yes") {
			my @ipComponents = split(/\//,$poolMember->{'IPAddress'});
			my $dnatTo = $ipComponents[0];
			if (@ipComponents < 2 || $ipComponents[1] eq "32" || $ipComponents[1] eq "128") {
				$changeSet->add([
						"/sbin/ip${ipv}tables",
							'-t','nat',
							'-A',"ots-dnat-$rxInterfaceID",
							'-d',$poolMember->{'IPNATAddress'},
							'-m','comment',
								'--comment', "pool: ".$pool->{'Name'}.", member: ".$poolMember->{'Username'},
							'-j','DNAT',
								'--to-destination', $dnatTo,
				]);
			} else {
				$logger->log(LOG_WARN,"[TC] Cannot add inbound NAT for pool member '%s' [%s] with IP '%s', NAT '%s' (inbound: %s) to pool '%s' [%s]",
						$poolMember->{'Username'},
						$poolMember->{'ID'},
						$poolMember->{'IPAddress'},
						$poolMember->{'IPNATAddress'} // "",
						$poolMember->{'IPNATInbound'} // "",
						$pool->{'Name'},
						$pool->{'ID'}
				);
			}
		}
	}


	# Post changeset
	$kernel->post("_tc" => "queue" => $changeSet);

	# Mark pool member as live
	unsetPoolMemberShaperState($poolMember->{'ID'},SHAPER_NOTLIVE|SHAPER_PENDING);
	setPoolMemberShaperState($poolMember->{'ID'},SHAPER_LIVE);
}



# Event handler for removing a pool member
sub _session_poolmember_remove
{
	my ($kernel, $pmid) = @_[KERNEL, ARG0];


	# Pull in pool member
	my $poolMember;
	if (!defined($poolMember = getPoolMember($pmid))) {
		$logger->log(LOG_ERR,"[TC] Shaper 'remove' event with non existing pool member '%s'",$pmid);
		return;
	}

	# Grab the pool members associated pool
	my $pool = getPool($poolMember->{'PoolID'});

	# Make sure its not NOTLIVE
	if (getPoolMemberShaperState($pmid) & SHAPER_NOTLIVE) {
		$logger->log(LOG_WARN,"[TC] Ignoring remove for pool member '%s' with IP '%s', NAT '%s' (inbound: %s) [%s] from pool '%s'",
				$poolMember->{'Username'},
				$poolMember->{'IPAddress'},
				$poolMember->{'IPNATAddress'} // "",
				$poolMember->{'IPNATInbound'} // "",
				$poolMember->{'ID'},
				$pool->{'Name'}
		);
		return;
	}

	$logger->log(LOG_INFO,"[TC] Remove pool member '%s' [%s] with IP '%s', NAT '%s' (inbound: %s) from pool '%s' [%s]",
			$poolMember->{'Username'},
			$poolMember->{'ID'},
			$poolMember->{'IPAddress'},
			$poolMember->{'IPNATAddress'} // "",
			$poolMember->{'IPNATInbound'} // "",
			$pool->{'Name'},
			$pool->{'ID'}
	);


	my $changeSet = TC::ChangeSet->new();

	my $txInterfaceID = getPoolTxInterface($pool->{'ID'});
	my $rxInterfaceID = getPoolRxInterface($pool->{'ID'});

	my $txInterface = getInterface($txInterfaceID);
	my $rxInterface = getInterface($rxInterfaceID);

	my $rxPoolTcClass = getPoolAttribute($pool->{'ID'},'tc.rxclass');
	my $txPoolTcClass = getPoolAttribute($pool->{'ID'},'tc.txclass');

	# Check what IP version we're dealing with
	my $ipv = "";
	if ( $poolMember->{'IPAddress'} =~ /:/ ) {
		$ipv = 6;
	}
	# Remove traffic classification
	$changeSet->add([
			"/sbin/ip${ipv}tables",
				'-t','mangle',
				'-D',"ots-tcfor-$rxInterfaceID",
				'-s',$poolMember->{'IPAddress'},
				'-m','comment',
					'--comment', "pool: ".$pool->{'Name'}.", member: ".$poolMember->{'Username'},
				'-j','CLASSIFY',
					'--set-class',"1:$rxPoolTcClass",
	]);
	$changeSet->add([
			"/sbin/ip${ipv}tables",
				'-t','mangle',
				'-D',"ots-tcfor-$txInterfaceID",
				'-d',$poolMember->{'IPAddress'},
				'-m','comment',
					'--comment', "pool: ".$pool->{'Name'}.", member: ".$poolMember->{'Username'},
				'-j','CLASSIFY',
					'--set-class',"1:$txPoolTcClass",
	]);

	# Remove NAT if we have any
	if (defined($poolMember->{'IPNATAddress'}) && defined($poolMember->{'IPNATAddress'}) ne "") {
		$changeSet->add([
				"/sbin/ip${ipv}tables",
					'-t','nat',
					'-D',"ots-snat-$txInterfaceID",
					'-s',$poolMember->{'IPAddress'},
					'-m','comment',
						'--comment', "pool: ".$pool->{'Name'}.", member: ".$poolMember->{'Username'},
					'-j','SNAT',
						'--to-source', $poolMember->{'IPNATAddress'},
		]);
		if (defined($poolMember->{'IPNATInbound'}) && $poolMember->{'IPNATInbound'} eq "yes") {
			my @ipComponents = split(/\//,$poolMember->{'IPAddress'});
			my $dnatTo = $ipComponents[0];
			if (@ipComponents < 2 || $ipComponents[1] eq "32" || $ipComponents[1] eq "128") {
				$changeSet->add([
						"/sbin/ip${ipv}tables",
							'-t','nat',
							'-D',"ots-dnat-$rxInterfaceID",
							'-d',$poolMember->{'IPNATAddress'},
							'-m','comment',
								'--comment', "pool: ".$pool->{'Name'}.", member: ".$poolMember->{'Username'},
							'-j','DNAT',
								'--to-destination', $dnatTo,
				]);
			} else {
				$logger->log(LOG_WARN,"[TC] Cannot remove inbound NAT for pool member '%s' [%s] with IP '%s', NAT '%s' (inbound: %s) to pool '%s' [%s]",
						$poolMember->{'Username'},
						$poolMember->{'ID'},
						$poolMember->{'IPAddress'},
						$poolMember->{'IPNATAddress'} // "",
						$poolMember->{'IPNATInbound'} // "",
						$pool->{'Name'},
						$pool->{'ID'}
				);
			}
		}
	}

	# Post changeset
	$kernel->post("_tc" => "queue" => $changeSet);

	# Mark as not live
	unsetPoolMemberShaperState($poolMember->{'ID'},SHAPER_LIVE|SHAPER_PENDING);
	setPoolMemberShaperState($poolMember->{'ID'},SHAPER_NOTLIVE);
}



# Grab pool ID from TC class
sub getPIDFromTcClass
{
	my ($interfaceID,$majorTcClass,$minorTcClass) = @_;


	# Return the pool ID if found
	my $ref = __getRefByMinorTcClass($interfaceID,$majorTcClass,$minorTcClass);
	if (!defined($ref) || substr($ref,0,13) ne "_pool_class_:") {
		return;
	}

	return substr($ref,13);
}



# Function to return if this is linked to a pool's class
sub isPoolTcClass
{
	my ($interfaceID,$majorTcClass,$minorTcClass) = @_;


	my $pid = getPIDFromTcClass($interfaceID,$majorTcClass,$minorTcClass);
	if (!defined($pid)) {
		return;
	}

	return $minorTcClass;
}



# Return the ClassID from a TC class
# This is similar to isTcTrafficClassValid() but returns the ref, not the minor class
sub getCIDFromTcClass
{
	my ($interfaceID,$majorTcClass,$minorTcClass) = @_;


	# Grab ref
	my $ref = __getRefByMinorTcClass($interfaceID,$majorTcClass,$minorTcClass);

	# If we're undefined return
	if (!defined($ref)) {
		return;
	}

	# If we're not a traffic class, just return
	if (substr($ref,0,16) ne "_traffic_class_:") {
		return;
	}

	# Else return the part after the above tag
	return substr($ref,16);
}


#
# Internal functions
#


# Function to initialize an interface
sub _tc_iface_init
{
	my ($changeSet,$interfaceID) = @_;


	# Grab our interface rate
	my $interface = getInterface($interfaceID);

### --- Interface Setup

	# Clear the qdisc from the interface
	$changeSet->add([
			'/sbin/tc','qdisc','del',
				'dev',$interface->{'Device'},
				'root',
	]);

	# Initialize the major TC class
	my $interfaceTcClass = _reserveMajorTcClass($interfaceID,"root");

	# Set interface RootClass
	$globals->{'Interfaces'}->{$interfaceID} = {
		'TcClass' => $interfaceTcClass
	};

### --- Interface Traffic Class Setup

	# Reserve our parent TC classes
	my @trafficClasses = getAllTrafficClasses();
	foreach my $trafficClassID (sort {$a <=> $b} @trafficClasses) {
		# Record the class we get for this interface traffic class ID
		my $interfaceTrafficClassTcClass = _reserveMinorTcClassByTrafficClassID($interfaceID,$trafficClassID);
	}

### --- Interface Setup Part 2

	# Add root qdisc
	$changeSet->add([
			'/sbin/tc','qdisc','add',
				'dev',$interface->{'Device'},
				'root',
				'handle','1:',
				'htb'
	]);

	# Attach our main limit on the qdisc
	my $burst = int(  ($interface->{'Limit'} / 8) * 1024 * 10); # Allow the entire interface to be emptied with a burst
	$changeSet->add([
			'/sbin/tc','class','add',
				'dev',$interface->{'Device'},
				'parent','1:',
				'classid','1:1',
				'htb',
					'rate',"$interface->{'Limit'}kbit",
	]);

	# Class 0 is our interface, it points to 1 (the major TcClass)) : 1 (class below)
	$globals->{'Interfaces'}->{$interfaceID}->{'TrafficClasses'}->{'0'} = {
		'TcClass' => '1',
		'CIR' => $interface->{'Limit'},
		'Limit' => $interface->{'Limit'}
	};


### --- Setup each class

	# Setup the classes
	foreach my $trafficClassID (@trafficClasses) {
		my $interfaceTrafficClassID = isInterfaceTrafficClassValid($interfaceID,$trafficClassID);
		my $interfaceTrafficClass = getEffectiveInterfaceTrafficClass2($interfaceTrafficClassID);
		my $tcClass = _getTcClassFromTrafficClassID($interfaceID,$trafficClassID);
		my $trafficPriority = getTrafficClassPriority($trafficClassID);

		# Add class
		$changeSet->add([
				'/sbin/tc','class','add',
					'dev',$interface->{'Device'},
					'parent','1:1',
					'classid',"1:$tcClass",
					'htb',
						'rate',"$interfaceTrafficClass->{'CIR'}kbit",
						'ceil',"$interfaceTrafficClass->{'Limit'}kbit",
						'prio',$trafficPriority,
		]);

		# Setup interface traffic class details
		$globals->{'Interfaces'}->{$interfaceID}->{'TrafficClasses'}->{$trafficClassID} = {
			'TcClass' => $tcClass,
			'CIR' => $interfaceTrafficClass->{'CIR'},
			'Limit' => $interfaceTrafficClass->{'Limit'}
		};
	}

	# Process our default pool traffic optimizations
	my $defaultPool = getInterfaceDefaultPool($interfaceID);
	if (defined($defaultPool)) {
		my $interfaceTrafficClassID = isInterfaceTrafficClassValid($interfaceID,$defaultPool);
		my $interfaceTrafficClass = getEffectiveInterfaceTrafficClass2($interfaceTrafficClassID);

		my $defaultPoolTcClass = _getTcClassFromTrafficClassID($interfaceID, $defaultPool);

		# Loop with IP versions
		for my $ipv ("", "6") {
			$changeSet->add([
					"/sbin/ip${ipv}tables",
						'-t','mangle',
						'-A',"ots-tcfor-$interfaceID",
						'-m','comment',
							'--comment', "Default",
						'-j','CLASSIFY',
							'--set-class',"1:$defaultPoolTcClass",
			]);
		}

		# If we have a rate for this iface, then use it
		_tc_class_optimize($changeSet,$interfaceID,$defaultPoolTcClass,$interfaceTrafficClass->{'Limit'});
	}
}



# Function to apply traffic optimizations to a classes
# XXX: This probably needs working on
sub _tc_class_optimize
{
	my ($changeSet,$interfaceID,$poolTcClass,$rate) = @_;


	my $interface = getInterface($interfaceID);
	my $prioTcClass = _reserveMajorTcClassByPrioClass($interfaceID,$poolTcClass);

	# Mbyte/s rate divided by average packet size to get packet limit, with a minimum of 127
	my $sfqQueueLength = max(int((($rate / 8) * 1024) / 1000), 127);
	# Flows is Mbit/s divided by 20, with minimum of 127
	my $sfqFlows = max(int($rate / 1000 / 20) * 127, 127);
	my $sfqRedFlowLimit = ($sfqQueueLength / 4) * 1500;  # Take 25% of the queue length and packet size of 1500

	# We then prioritize traffic into 3 bands based on TOS
	$changeSet->add([
			'/sbin/tc','qdisc','add',
				'dev',$interface->{'Device'},
				'parent',"1:$poolTcClass",
				'handle',"$prioTcClass:",
				'sfq',
						'limit', $sfqQueueLength,  # first work out byte rate, then divide by average packet size of 1000
						'flows', $sfqFlows,
						'headdrop',
						'perturb', 10,
						'redflowlimit', $sfqRedFlowLimit,
						'perturb', 10,
	]);

}



# Function to add a TC class
sub _tc_class_add
{
	my ($changeSet,$interfaceID,$majorTcClass,$trafficClassTcClass,$poolTcClass,$rate,$ceil,$trafficPriority) = @_;


	my $interface = getInterface($interfaceID);

	# # Set burst to a sane value, in this case the CIR (or $rate) size
	# my $burst = int($rate / 8);
	# my $cburst = int(($ceil - $rate) / 8 / 10);
	# # NK: cburst should not exceed burst, if it does, just use the burst value
	# # this ensures we do not get negative burst
	# if ($cburst > $burst) {
	# 	$cburst = $burst;
	# }
	my $burst = int( ($ceil / 8) * 1024 * 10);
	my $cburst = int($rate / 8 / 5) + 2; # +2kb to cover 1600
	my $quantum = int($cburst / 2) * 1024; # Burst in bytes

	# Create main rate limiting classes
	$changeSet->add([
			'/sbin/tc','class','add',
				'dev',$interface->{'Device'},
				'parent',"$majorTcClass:$trafficClassTcClass",
				'classid',"$majorTcClass:$poolTcClass",
				'htb',
					'rate', "${rate}kbit",
					'ceil', "${ceil}kbit",
#					'prio', $trafficPriority,
#					'burst', "${burst}kb",
#					'cburst', "${cburst}kb",
	]);
}



# Function to change a TC class
sub _tc_class_change
{
	my ($changeSet,$interfaceID,$majorTcClass,$trafficClassTcClass,$poolTcClass,$rate,$ceil,$trafficPriority) = @_;


	my $interface = getInterface($interfaceID);

	my @args = ();

	# If ceil is not available, set it to the CIR (or $rate in this case)
	if (!defined($ceil)) {
		$ceil = $rate;
	}

	# # Set the burst rate to the CIR (or $rate in this case)
	# my $burst = int($rate / 8);
	# my $cburst = int(($ceil - $rate) / 8 / 10);
	# # NK: cburst should not exceed burst, if it does, just use the burst value
	# # this ensures we do not get negative burst
	# if ($cburst > $burst) {
	# 	$cburst = $burst;
	# }
	my $burst = int( ($ceil / 8) * 1024 * 10);
	my $cburst = int($rate / 8 / 5) + 2; # +2kb to cover 1600
	my $quantum = int($cburst / 2) * 1024; # Burst in bytes

	# # Check if we have a priority
	# if (defined($trafficPriority)) {
	# 	push(@args,'prio',$trafficPriority);
	# }

	# Create main rate limiting classes
	$changeSet->add([
			'/sbin/tc','class','change',
				'dev',$interface->{'Device'},
				'parent',"$majorTcClass:$trafficClassTcClass",
				'classid',"$majorTcClass:$poolTcClass",
				'htb',
					'rate', "${rate}kbit",
					'ceil', "${ceil}kbit",
#					'burst', "${burst}kb",
#					'cburst', "${cburst}kb",
					@args
	]);
}



# Get a pool TC class from pool ID
sub _reserveMinorTcClassByPoolID
{
	my ($interfaceID,$pid) = @_;

	return __reserveMinorTcClass($interfaceID,TC_ROOT_CLASS,"_pool_class_:$pid");
}



# Get a traffic class TC class
sub _reserveMinorTcClassByTrafficClassID
{
	my ($interfaceID,$trafficClassID) = @_;

	return __reserveMinorTcClass($interfaceID,TC_ROOT_CLASS,"_traffic_class_:$trafficClassID");
}



# Get a prio class TC class
# This is a MAJOR class!
sub _reserveMajorTcClassByPrioClass
{
	my ($interfaceID,$trafficClassID) = @_;


	return _reserveMajorTcClass($interfaceID,"_priority_class_:$trafficClassID");
}



# Return TC class from a traffic class ID
sub _getTcClassFromTrafficClassID
{
	my ($interfaceID,$trafficClassID) = @_;


	return __getMinorTcClassByRef($interfaceID,TC_ROOT_CLASS,"_traffic_class_:$trafficClassID");
}



# Return prio TC class using class
# This returns a MAJOR class from a tc class
sub _getPrioTcClass
{
	my ($interfaceID,$tcClass) = @_;

	return __getMajorTcClassByRef($interfaceID,"_priority_class_:$tcClass");
}



# Function to dispose of a TC class
sub _disposePoolTcClass
{
	my ($interfaceID,$tcClass) = @_;

	return __disposeMinorTcClass($interfaceID,TC_ROOT_CLASS,$tcClass);
}



# Function to dispose of a major TC class
# Uses a TC class to get a MAJOR class, then disposes it
sub _disposePrioTcClass
{
	my ($interfaceID,$tcClass) = @_;


	# If we can grab the major class dipose of it
	my $majorTcClass = _getPrioTcClass($interfaceID,$tcClass);
	if (!defined($majorTcClass)) {
		return;
	}

	return __disposeMajorTcClass($interfaceID,$majorTcClass);
}



# Function to get next available TC class
sub __reserveMinorTcClass
{
	my ($interfaceID,$majorTcClass,$ref) = @_;


	# Setup defaults if we don't have anything defined
	if (!defined($globals->{'TcClasses'}->{$interfaceID}) || !defined($globals->{'TcClasses'}->{$interfaceID}->{$majorTcClass})) {
		$globals->{'TcClasses'}->{$interfaceID}->{$majorTcClass} = {
			# Skip 0 and 1
			'Counter' => 2,
			'Free' => [ ],
			'Track' => { },
			'Reverse' => { },
		};
	}

	# Maybe we have one free?
	my $minorTcClass = shift(@{$globals->{'TcClasses'}->{$interfaceID}->{$majorTcClass}->{'Free'}});

	# Generate new number
	if (!$minorTcClass) {
		$minorTcClass = $globals->{'TcClasses'}->{$interfaceID}->{$majorTcClass}->{'Counter'}++;
		# Hex it
		$minorTcClass = toHex($minorTcClass);
	}

	$globals->{'TcClasses'}->{$interfaceID}->{$majorTcClass}->{'Track'}->{$minorTcClass} = $ref;
	$globals->{'TcClasses'}->{$interfaceID}->{$majorTcClass}->{'Reverse'}->{$ref} = $minorTcClass;

	return $minorTcClass;
}



# Function to get next available major TC class
sub _reserveMajorTcClass
{
	my ($interfaceID,$ref) = @_;


	# Setup defaults if we don't have anything defined
	if (!defined($globals->{'TcClasses'}->{$interfaceID})) {
		$globals->{'TcClasses'}->{$interfaceID} = {
			# Skip 0
			'Counter' => 1,
			'Free' => [ ],
			'Track' => { },
			'Reverse' => { },
		};
	}

	# Maybe we have one free?
	my $majorTcClass = shift(@{$globals->{'TcClasses'}->{$interfaceID}->{'Free'}});

	# Generate new number
	if (!$majorTcClass) {
		$majorTcClass = $globals->{'TcClasses'}->{$interfaceID}->{'Counter'}++;
		# Hex it
		$majorTcClass = toHex($majorTcClass);
	}

	$globals->{'TcClasses'}->{$interfaceID}->{'Track'}->{$majorTcClass} = $ref;
	$globals->{'TcClasses'}->{$interfaceID}->{'Reverse'}->{$ref} = $majorTcClass;

	return $majorTcClass;
}



# Get a minor class by its rerf
sub __getMinorTcClassByRef
{
	my ($interfaceID,$majorTcClass,$ref) = @_;


	if (!defined($globals->{'TcClasses'}->{$interfaceID}) || !defined($globals->{'TcClasses'}->{$interfaceID}->{$majorTcClass})) {
		return;
	}

	return $globals->{'TcClasses'}->{$interfaceID}->{$majorTcClass}->{'Reverse'}->{$ref};
}



# Get a major class by its rerf
sub __getMajorTcClassByRef
{
	my ($interfaceID,$ref) = @_;


	if (!defined($globals->{'TcClasses'}->{$interfaceID})) {
		return;
	}

	return $globals->{'TcClasses'}->{$interfaceID}->{'Reverse'}->{$ref};
}



# Get ref using the minor tc class
sub __getRefByMinorTcClass
{
	my ($interfaceID,$majorTcClass,$minorTcClass) = @_;


	if (!defined($globals->{'TcClasses'}->{$interfaceID}) || !defined($globals->{'TcClasses'}->{$interfaceID}->{$majorTcClass})) {
		return;
	}

	return $globals->{'TcClasses'}->{$interfaceID}->{$majorTcClass}->{'Track'}->{$minorTcClass};
}



# Function to dispose of a TC class
sub __disposeMinorTcClass
{
	my ($interfaceID,$majorTcClass,$tcMinorClass) = @_;


	my $ref = $globals->{'TcClasses'}->{$interfaceID}->{$majorTcClass}->{'Track'}->{$tcMinorClass};
	# Push onto free list
	push(@{$globals->{'TcClasses'}->{$interfaceID}->{$majorTcClass}->{'Free'}},$tcMinorClass);
	# Blank the value
	$globals->{'TcClasses'}->{$interfaceID}->{$majorTcClass}->{'Track'}->{$tcMinorClass} = undef;
	delete($globals->{'TcClasses'}->{$interfaceID}->{$majorTcClass}->{'Reverse'}->{$ref});
}



# Function to dispose of a major TC class
sub __disposeMajorTcClass
{
	my ($interfaceID,$tcMajorClass) = @_;


	my $ref = $globals->{'TcClasses'}->{$interfaceID}->{'Track'}->{$tcMajorClass};
	# Push onto free list
	push(@{$globals->{'TcClasses'}->{$interfaceID}->{'Free'}},$tcMajorClass);
	# Blank the value
	$globals->{'TcClasses'}->{$interfaceID}->{'Track'}->{$tcMajorClass} = undef;
	delete($globals->{'TcClasses'}->{$interfaceID}->{'Reverse'}->{$ref});
}



# Function to get next available TC filter
sub _reserveTcFilter
{
	my ($interfaceID,$ref) = @_;


	# Setup defaults if we don't have anything defined
	if (!defined($globals->{'TcFilters'}->{$interfaceID})) {
		$globals->{'TcFilters'}->{$interfaceID} = {
			# Skip 0 and 1
			'Counter' => 2,
			'Free' => [ ],
			'Track' => { },
		};
	}

	# Maybe we have one free?
	my $filterID = shift(@{$globals->{'TcFilters'}->{$interfaceID}->{'Free'}});

	# Generate new number
	if (!$filterID) {
		$filterID = $globals->{'TcFilters'}->{$interfaceID}->{'Counter'}++;
		# We cannot use ID 800, its internal
		$filterID = $globals->{'TcFilters'}->{$interfaceID}->{'Counter'}++ if ($filterID == 800);
		# Hex it
		$filterID = toHex($filterID);
	}

	$globals->{'TcFilters'}->{$interfaceID}->{'Track'}->{$filterID} = $ref;

	return $filterID;
}



# Function to dispose of a TC Filter
sub _disposeTcFilter
{
	my ($interfaceID,$filterID) = @_;

	# Push onto free list
	push(@{$globals->{'TcFilters'}->{$interfaceID}->{'Free'}},$filterID);
	# Blank the value
	$globals->{'TcFilters'}->{$interfaceID}->{'Track'}->{$filterID} = undef;
}



#
# Task/child communication & handling stuff
#



# Initialize our tc session
sub _task_session_start
{
	my $kernel = $_[KERNEL];

	# Set our alias
	$kernel->alias_set("_tc");

	# Setup handing of console INT
	$kernel->sig("INT", "_SIGINT");

	# Fire things up, we trigger this to process the task queue generated during init
	$kernel->yield("_task_run_next");
}



# Add task to queue
sub _task_add_to_queue
{
	my $changeSet = shift;


	# Extract the changeset into commands
	my $numChanges = 0;
	foreach my $cmd ($changeSet->extract()) {
		push(@{$globals->{'TaskQueue'}},$cmd);
		$numChanges++;
	}

	$logger->log(LOG_DEBUG,"[TC] TASK: Queued %s changes",$numChanges);
}



# Queue a task
sub _task_queue
{
	my ($kernel,$heap,$changeSet) = @_[KERNEL,HEAP,ARG0];


	# Internal function to add command to queue
	_task_add_to_queue($changeSet);

	# Trigger a run if list is not empty
	if (@{$globals->{'TaskQueue'}}) {
		$kernel->yield("_task_run_next");
	}
}



# Run next task
sub _task_run_next
{
	my ($kernel,$heap) = @_[KERNEL,HEAP];

	if (keys %{$heap->{'task_by_wid'}}) {
		# NK: Limit concurrency to 1
		return;
	}

	# Check if we have a task coming off the top of the task queue
	if (my $cmd = shift(@{$globals->{'TaskQueue'}})) {
		my $cmdStr = encode_json($cmd);

		# Create task
		my $task = POE::Wheel::Run->new(
			Program => $cmd,
			StdoutFilter => POE::Filter::Line->new( Literal => "\n" ),
			StderrFilter => POE::Filter::Line->new( Literal => "\n" ),
			StdoutEvent => '_task_child_stdout',
			StderrEvent => '_task_child_stderr',
			CloseEvent => '_task_child_close',
		#	ErrorEvent => '_task_child_error',
		) or $logger->log(LOG_ERR,"[TC] TASK: Unable to start task: $cmdStr");

		# Set task ID
		my $task_id = $task->ID;

		# Intercept SIGCHLD
		$kernel->sig_child($task->PID, "_SIGCHLD");

		# Wheel events include the wheel's ID.
		$heap->{'task_by_wid'}->{$task_id} = $task;
		# Signal events include the process ID.
		$heap->{'task_by_pid'}->{$task->PID} = $task;

		# Build commandline string
		$logger->log( LOG_DEBUG, "[TC] TASK/%s: Starting '%s' as %s with PID %s", $task->ID, $cmdStr, $task->ID, $task->PID );
	}

}



# Child writes to STDOUT
sub _task_child_stdout
{
	my ($kernel,$heap,$stdout,$task_id) = @_[KERNEL,HEAP,ARG0,ARG1];


	my $task = $heap->{'task_by_wid'}->{$task_id};

	$logger->log(LOG_INFO,"[TC] TASK/%s: STDOUT => %s",$task_id,$stdout);
}



# Child writes to STDERR
sub _task_child_stderr
{
	my ($kernel,$heap,$stdout,$task_id) = @_[KERNEL,HEAP,ARG0,ARG1];


	my $task = $heap->{'task_by_wid'}->{$task_id};

	$logger->log(LOG_WARN,"[TC] TASK/%s: STDERR => %s",$task_id,$stdout);
}



# Child closed its handles, it won't communicate with us, so remove it
sub _task_child_close
{
	my ($kernel,$heap,$task_id) = @_[KERNEL,HEAP,ARG0];


	my $task = $heap->{'task_by_wid'}->{$task_id};

	# May have been reaped by task_sigchld()
	if (!defined($task)) {
		$logger->log(LOG_DEBUG,"[TC] TASK/%s: Closed dead child",$task_id);
		return;
	}

	$logger->log(LOG_DEBUG,"[TC] TASK/%s: Closed PID %s",$task_id,$task->PID);

	# Remove other references
	delete($heap->{'task_by_wid'}->{$task_id});
	delete($heap->{'task_by_pid'}->{$task->PID});

	# Start next one, if there is a next one
	if (@{$globals->{'TaskQueue'}}) {
		$kernel->yield("_task_run_next");
	}
}



# Child got an error event, lets remove it too
sub _task_child_error
{
	my ($kernel,$heap,$operation,$errnum,$errstr,$task_id) = @_[KERNEL,HEAP,ARG0..ARG3];


	my $task = $heap->{'task_by_wid'}->{$task_id};

	if ($operation eq "read" && !$errnum) {
		$errstr = "Remote end closed"
	}

	$logger->log(LOG_ERR,"[TC] Task %s generated %s error %s: '%s'",$task_id,$operation,$errnum,$errstr);

	# If there is no task, return
	return if (!defined($task));

	# Remove other references
	delete($heap->{'task_by_wid'}->{$task_id});
	delete($heap->{'task_by_pid'}->{$task->PID});

	# Start next one, if there is a next one
	if (@{$globals->{'TaskQueue'}}) {
		$kernel->yield("_task_run_next");
	}
}



# Reap the dead child
sub _task_SIGCHLD
{
	my ($kernel,$heap,$pid,$status) = @_[KERNEL,HEAP,ARG1,ARG2];


	my $task = $heap->{'task_by_pid'}->{$pid};

	$logger->log(LOG_DEBUG,"[TC] TASK: Task with PID %s exited with status %s",$pid,$status);

	# May have been reaped by task_child_close()
	return if (!defined($task));

	# Remove other references
	delete($heap->{'task_by_wid'}->{$task->ID});
	delete($heap->{'task_by_pid'}->{$pid});
}



# Handle SIGINT
sub _task_SIGINT
{
	my ($kernel,$heap,$signal_name) = @_[KERNEL,HEAP,ARG0];


	# Shutdown stdin on all children, this will terminate /sbin/tc
	foreach my $task_id (keys %{$heap->{'task_by_wid'}}) {
		my $task = $heap->{'task_by_wid'}{$task_id};
#		$kernel->sig_child($task->PID, "asig_child");
#		$task->kill("INT"); #NK: doesn't work
		$kernel->post($task,"shutdown_stdin"); #NK: doesn't work
	}

	$logger->log(LOG_WARN,"[TC] Killed children processes");
}



# TC changeset item
package TC::ChangeSet;

use strict;
use warnings;

# Create object
sub new
{
	my $class = shift;

	my $self = {
		'list' => [ ]
	};

	bless $self, $class;
	return $self;
}



# Add a change to the list
sub add
{
	my ($self,$change) = @_;

	push(@{$self->{'list'}},$change);
}



# Return the list
sub extract
{
	my $self = shift;

	return @{$self->{'list'}};
}



# Return the list
sub debug
{
	my $self = shift;

	my @debug = ();
	foreach my $item ($self->extract) {
		push(@debug,join(' ',@{$item}));
	}

	return @debug;
}



1;
# vim: ts=4
